import argparse
import numpy as np
import torch
import random
import os
import json
import datetime
from torch.utils.tensorboard import SummaryWriter

from src.stats import Statistics
from src.buffer import ReplayBuffer
from src.cql_sac.agent import CSCCQLSAC
from src.environment import create_env

##################
# ARGPARSER
##################
def cmd_args():
    parser = argparse.ArgumentParser(formatter_class=lambda prog:argparse.ArgumentDefaultsHelpFormatter(prog, max_help_position=40))
    # environment args
    env_args = parser.add_argument_group('Environment')
    env_args.add_argument("--env_id", action="store", type=str, default="SafetyPointGoal1-v0", metavar="ID",
                        help="Set the environment")
    env_args.add_argument('--env_args', type=str, default='',
                    help='Environment specific arguments')
    env_args.add_argument("--cost_limit", action="store", type=float, default=0.0, metavar="N",
                        help="Set a cost limit/budget")

    # train and test args
    train_test_args = parser.add_argument_group('Train and Test')
    train_test_args.add_argument("--total_train_steps", action="store", type=int, default=5_000_000, metavar="N",
                    help="Total number of steps until training is finished")
    train_test_args.add_argument("--train_episodes", action="store", type=int, default=4, metavar="N",
                        help="Number of episodes until policy optimization")
    train_test_args.add_argument("--train_until_test", action="store", type=int, default=4, metavar="N",
                        help="Perform evaluation after N * total_train_episodes episodes of training")
    train_test_args.add_argument("--update_iterations", action="store", type=int, default=64, metavar="N",
                        help="Number of updates performed after each training step")
    train_test_args.add_argument("--test_episodes", action="store", type=int, default=16, metavar="N",
                        help="Number of episodes used for testing")
    
    # update args
    update_args = parser.add_argument_group('Update')
    update_args.add_argument("--batch_size", action="store", type=int, default=256, metavar="N",
                        help="Batch size used for training")
    update_args.add_argument("--tau", action="store", type=float, default=0.005, metavar="N",
                        help="Factor used in soft update of target network")
    update_args.add_argument("--gamma", action="store", type=float, default=0.95, metavar="N",
                        help="Discount factor for rewards")
    update_args.add_argument("--learning_rate", action="store", type=float, default=0.0003, metavar="N",
                        help="Learn rate for the policy and Q networks")
    
    # buffer args
    buffer_args = parser.add_argument_group('Buffer')
    buffer_args.add_argument("--buffer_capacity", action="store", type=int, default=250_000, metavar="N",
                        help="Define the maximum capacity of the replay buffer")
    buffer_args.add_argument("--clear_buffer", action="store_true", default=False,
                        help="Clear Replay Buffer after every optimization step")
    
    # network args
    network_args = parser.add_argument_group('Networks')
    network_args.add_argument("--hidden_size", action="store", type=int, default=256, metavar="N",
                        help="Hidden size of the networks")
    
    # cql args
    cql_args = parser.add_argument_group('CQL')
    cql_args.add_argument("--cql_with_lagrange", action="store_true", default=False,
                        help="Enable automatic alpha tuning")
    cql_args.add_argument("--cql_temp", action="store", type=float, default=1.0, metavar="N",
                        help="Set the temperature for the CQL")
    cql_args.add_argument("--cql_weight", action="store", type=float, default=1.0, metavar="N",
                        help="Set the weight for the CQL")
    cql_args.add_argument("--cql_target_action_gap", action="store", type=float, default=10, metavar="N",
                        help="Set the target action gap for the CQL")
    
    # csc args
    csc_args = parser.add_argument_group('CSC')
    csc_args.add_argument("--csc_chi", action="store", type=float, default=0.05, metavar="N",
                        help="Set the value of chi")
    csc_args.add_argument("--csc_delta", action="store", type=float, default=0.01, metavar="N",
                        help="Set the value of delta")
    csc_args.add_argument("--csc_beta", action="store", type=float, default=0.7, metavar="N",
                        help="Set the value of beta")
    csc_args.add_argument("--csc_alpha", action="store", type=float, default=0.5, metavar="N",
                        help="Set the value of alpha")
    csc_args.add_argument("--csc_lambda", action="store", type=float, default=1.0, metavar="N",
                        help="Set the initial value of lambda")
    csc_args.add_argument("--csc_shield_iterations", action="store", type=int, default=100, metavar="N",
                        help="Set the number of sampled actions during shielding")
    csc_args.add_argument("--csc_update_iterations", action="store", type=int, default=20, metavar="N",
                        help="Set the number of linear backtracking iterations")

    # common args
    common_args = parser.add_argument_group('Common')
    common_args.add_argument("--seed", action="store", type=int, default=0, metavar="N",
                        help="Set a custom seed for the rng")
    common_args.add_argument("--device", action="store", type=str, default="cuda", metavar="DEVICE",
                        help="Set the device for pytorch to use")
    common_args.add_argument("--log_dir", action="store", type=str, default="./runs", metavar="PATH",
                        help="Set the output and log directory path")
    common_args.add_argument("--num_threads", action="store", type=int, default=1, metavar="N",
                        help="Set the maximum number of threads for pytorch and numpy")
    
    # dev args
    dev_args = parser.add_argument_group('Development')
    dev_args.add_argument("--fisher_inversion", action="store", type=str, default="pinv", metavar="INV",
                        help="Inversion mode for fisher matrix: pinv, inv, none")

    args = parser.parse_args()
    return args

##################
# SETUP
##################

def setup(args):
    """
    Performs setup like fixing seeds, initializing env and agent, buffer and stats.
    """
    torch.set_num_threads(args.num_threads)
    torch.set_default_dtype(torch.double)
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)

    output_dir = os.path.join(args.log_dir, datetime.datetime.now().strftime("%d_%m_%y__%H_%M_%S"))
    writer = SummaryWriter(log_dir=output_dir)
    with open(os.path.join(output_dir, "config.json"), "w") as file:
        json.dump(args.__dict__, file, indent=2)

    env = create_env(args=args)
    buffer = ReplayBuffer(env=env, cap=args.buffer_capacity)
    stats = Statistics(writer=writer)
    agent = CSCCQLSAC(env=env, args=args, stats=stats)

    return env, agent, buffer, stats

##################
# EXPLORATION
##################

@torch.no_grad
def run_single_exploration(args, env, agent:CSCCQLSAC, buffer:ReplayBuffer, stats:Statistics, train=True, shielded=True):
    # track leftover episodes
    num_episodes = args.train_episodes if train else args.test_episodes

    avg_steps = 0.
    avg_return = 0.
    avg_cost = 0.
    avg_unsafe = 0.
    
    for _ in range(num_episodes):
        # initialize stats per episode
        episode_steps = 0
        episode_reward = 0.
        episode_cost = 0.
        done = False

        # reset env
        state, info = env.reset(seed=random.randint(0, 2**31-1))

        while not done:
            # sample and execute actions
            action = agent.get_action(state, shielded=shielded).cpu().numpy()
            next_state, reward, cost, terminated, truncated, info = env.step(action)
            done = terminated or truncated

            # increment stats
            episode_steps += 1
            episode_reward += reward
            episode_cost += cost

            # During training, add experiences to buffer
            if train:
                buffer.add(
                    state, action, reward, cost, next_state, terminated
                )
            
            # End episode if the cost exceeded the limit
            if episode_cost > args.cost_limit:
                done = True
            
            # Update state
            state = next_state

        # Update statistics
        count_unsafe = int(episode_cost > args.cost_limit)
        stats.total_train_episodes += 1
        stats.total_train_steps += episode_steps
        stats.total_train_unsafe += count_unsafe

        # record finished episodes
        stats.log_train_tensorboard(
            episode_steps=episode_steps,
            episode_return=episode_reward,
            episode_cost=episode_cost
        )
        
        avg_return += episode_reward
        avg_steps += episode_steps
        avg_cost += episode_cost
        avg_unsafe += count_unsafe

    # average and log stats
    avg_return /= num_episodes
    avg_steps /= num_episodes
    avg_cost /= num_episodes
    avg_unsafe /= num_episodes

    # In training, update avg_unsafe
    # In testing, log averages
    if train:
        stats.train_unsafe_avg = avg_unsafe
    else:
        stats.log_test_tensorboard(
            avg_steps=avg_steps,
            avg_return=avg_return,
            avg_cost=avg_cost,
            avg_unsafe=avg_unsafe,
            shielded=shielded
        )

##################
# MAIN LOOP
##################

def main(args, env, agent, buffer, stats):
    finished = False
    while not finished:
        # Training + Update Loop
        for _ in range(args.train_until_test):

            # Check if training is finished
            if stats.total_train_steps >= args.total_train_steps:
                finished = True
                break

            # 1. Run exploration for training
            run_single_exploration(args, env, agent, buffer, stats, train=True, shielded=True)

            # 2. Perform updates
            for _ in range(args.update_iterations):
                experiences=buffer.sample(n=args.batch_size)
                agent.learn(experiences=experiences)

            # 3. After update stuff
            if args.clear_buffer:
                buffer.clear()

        # Test loop (shielded and unshielded)
        for shielded in [True, False]:
            run_single_exploration(args, env, agent, buffer, stats, train=False, shielded=shielded)

if __name__ == '__main__':
    args = cmd_args()
    main(args, *setup(args))