import numpy as np

class ReplayBuffer():
    """
    Buffer for storing experiences. Supports sampling and adding experiences and clearing the buffer.
    """
    def __init__(self, env, cap):
        self._cap = max(1,cap)
        self._size = 0              # number of experiences in the buffer
        self._ptr = 0               # pointer to the next available slot in the buffer

        self._states = np.zeros((cap, env.observation_space.shape[0]), dtype=np.float64)
        self._actions = np.zeros((cap, env.action_space.shape[0]), dtype=np.float64)
        self._rewards = np.zeros((cap, ), dtype=np.float64)
        self._costs = np.zeros((cap, ), dtype=np.float64)
        self._next_states = np.zeros_like(self._states)
        self._dones = np.zeros((cap, ), dtype=np.uint8)

    def _add(self, state, action, reward, cost, next_state, done):
        idx = self._ptr
        self._states[idx] = state
        self._actions[idx] = action
        self._rewards[idx] = reward
        self._costs[idx] = cost
        self._next_states[idx] = next_state
        self._dones[idx] = done

    def add(self, state, action, reward, cost, next_state, done):
        """
        Adds an experience to the buffer.
        """
        self._add(state, action, reward, cost, next_state, done)
        self._ptr += 1
        self._size = max(self._size, self._ptr)
        self._ptr = self._ptr % self._cap   

    def sample(self, n):
        """
        Samples n experiences from the buffer.
        """
        idxs = np.random.randint(low=0, high=self._size, size=n)
        return self._states[idxs], self._actions[idxs], self._rewards[idxs], \
            self._costs[idxs], self._next_states[idxs], self._dones[idxs]
    
    def clear(self):
        """
        Clears the buffer.
        """
        self._size = 0
        self._ptr = 0