import torch
import torch.optim as optim
import torch.nn.functional as F
import torch.nn as nn
from torch.nn.utils import clip_grad_norm_
from .networks import Critic, Actor
import math
import copy

class CSCCQLSAC(nn.Module):
    """Interacts with and learns from the environment."""
    
    def __init__(self,
                        env,
                        args,
                        stats
                ):
        """Initialize an Agent object.
        
        Params
        ======
            env : the vector environment
            args : the argparse arguments
        """
        super(CSCCQLSAC, self).__init__()
        self.stats = stats

        state_size = env.observation_space.shape[0]
        action_size = env.action_space.shape[0]
        hidden_size = args.hidden_size
        self.action_size = action_size

        self.device = args.device
        
        self.learning_rate = args.learning_rate
        self.gamma = args.gamma
        self.tau = args.tau
        self.clip_grad_param = 1

        self.target_entropy = -action_size  # -dim(A)

        self.log_alpha = torch.tensor([0.0], requires_grad=True, device=self.device)
        self.alpha = self.log_alpha.exp().detach()
        self.alpha_optimizer = optim.Adam(params=[self.log_alpha], lr=self.learning_rate) 
        
        # CQL params
        self.cql_with_lagrange = args.cql_with_lagrange
        self.cql_temp = args.cql_temp
        self.cql_weight = args.cql_weight
        self.cql_target_action_gap = args.cql_target_action_gap
        self.cql_log_alpha = torch.zeros(1, requires_grad=True, device=self.device)
        self.cql_alpha_optimizer = optim.Adam(params=[self.cql_log_alpha], lr=self.learning_rate)

        # CSC params
        self.csc_shield_iterations = args.csc_shield_iterations
        self.csc_update_iterations = args.csc_update_iterations
        self.csc_alpha = args.csc_alpha
        self.csc_beta = args.csc_beta
        self.csc_delta = args.csc_delta
        self.csc_chi = args.csc_chi
        self.csc_avg_unsafe = args.csc_chi

        self.csc_lambda = torch.tensor([args.csc_lambda], requires_grad=True, device=self.device)
        self.csc_lambda_optimizer = optim.Adam(params=[self.csc_lambda], lr=self.learning_rate)

        # Dev args
        self.fisher_inversion = args.fisher_inversion
        
        # Actor Network 
        self.actor_local = Actor(state_size, action_size, hidden_size).to(self.device)
        self.actor_optimizer = optim.Adam(self.actor_local.parameters(), lr=self.learning_rate)

        # Safety Critic Network (w/ Target Network)
        self.safety_critic1 = Critic(state_size, action_size, hidden_size).to(self.device)
        self.safety_critic2 = Critic(state_size, action_size, hidden_size).to(self.device)
        
        self.safety_critic1_target = Critic(state_size, action_size, hidden_size).to(self.device)
        self.safety_critic1_target.load_state_dict(self.safety_critic1.state_dict())

        self.safety_critic2_target = Critic(state_size, action_size, hidden_size).to(self.device)
        self.safety_critic2_target.load_state_dict(self.safety_critic2.state_dict())

        self.safety_critic1_optimizer = optim.Adam(self.safety_critic1.parameters(), lr=self.learning_rate)
        self.safety_critic2_optimizer = optim.Adam(self.safety_critic2.parameters(), lr=self.learning_rate)
        
        # Critic Network (w/ Target Network)
        self.critic1 = Critic(state_size, action_size, hidden_size).to(self.device)
        self.critic2 = Critic(state_size, action_size, hidden_size).to(self.device)
        
        self.critic1_target = Critic(state_size, action_size, hidden_size).to(self.device)
        self.critic1_target.load_state_dict(self.critic1.state_dict())

        self.critic2_target = Critic(state_size, action_size, hidden_size).to(self.device)
        self.critic2_target.load_state_dict(self.critic2.state_dict())

        self.critic1_optimizer = optim.Adam(self.critic1.parameters(), lr=self.learning_rate)
        self.critic2_optimizer = optim.Adam(self.critic2.parameters(), lr=self.learning_rate)

    def get_action(self, state, shielded=True):
        """ 
        Returns shielded actions for given state as per current policy.
        """
        state = torch.from_numpy(state).to(torch.get_default_dtype()).to(self.device).unsqueeze(0)
        
        if shielded:
            # Repeat state, resulting shape: (shield_iterations, state_size)
            state = state.repeat((self.csc_shield_iterations, 1))
            unsafety_threshold = (1 - self.gamma) * (self.csc_chi - self.csc_avg_unsafe)
            
            # Sample all 'csc_shield_iterations' actions at once
            with torch.no_grad():
                action = self.actor_local.get_action(state).to(self.device)
            
            # Estimate unsafety of all actions
            q1 = self.safety_critic1(state, action)
            q2 = self.safety_critic2(state, action)
            unsafety = torch.min(q1, q2).squeeze(1)

            # Check for actions that qualify (unsafety <= threshold), locate first (if exists)
            mask_safe = unsafety <= unsafety_threshold

            # Search for first safe action, check if one exists
            idx_first_safe = mask_safe.int().argmax(dim=0).item()
            is_safe = mask_safe[idx_first_safe].item()

            # Return first safe or alternatively the safest one
            if is_safe:
                return action[idx_first_safe, :]
            else:
                idx_best = unsafety.argmin(dim=0).item()
                return action[idx_best, :]
        else:
            return self.actor_local.get_action(state).squeeze(0).to(self.device)

    def calc_policy_loss(self, states, alpha):
        actions_pred, log_pis = self.actor_local.evaluate(states)

        q1 = self.critic1(states, actions_pred.squeeze(0))
        q2 = self.critic2(states, actions_pred.squeeze(0))
        min_Q = torch.min(q1,q2)
        actor_loss = ((alpha * log_pis - min_Q )).mean()
        return actor_loss, log_pis

    def _compute_policy_values(self, obs_pi, obs_q):
        #with torch.no_grad():
        actions_pred, log_pis = self.actor_local.evaluate(obs_pi)
        qs1 = self.safety_critic1(obs_q, actions_pred)
        qs2 = self.safety_critic2(obs_q, actions_pred)
        return qs1 - log_pis.detach(), qs2 - log_pis.detach()
    
    def _compute_random_values(self, obs, actions, critic):
        random_values = critic(obs, actions)
        random_log_probs = math.log(0.5 ** self.action_size)
        return random_values - random_log_probs
    
    def learn(self, experiences):
        """ Updates actor, critics, safety_critics and entropy_alpha parameters using given batch of experience tuples.
        Params
        ======
            experiences (Tuple[torch.Tensor]): tuple of (s, a, r, c, s', done) tuples
        """
        self.csc_avg_unsafe = self.stats.train_unsafe_avg
        states, actions, rewards, costs, next_states, dones = experiences

        states = torch.from_numpy(states).to(torch.get_default_dtype()).to(self.device)
        actions = torch.from_numpy(actions).to(torch.get_default_dtype()).to(self.device)
        rewards = torch.from_numpy(rewards).to(torch.get_default_dtype()).to(self.device).view(-1, 1)
        costs = torch.from_numpy(costs).to(torch.get_default_dtype()).to(self.device).view(-1, 1)
        next_states = torch.from_numpy(next_states).to(torch.get_default_dtype()).to(self.device)
        dones = torch.from_numpy(dones).to(torch.get_default_dtype()).to(self.device).view(-1, 1)

        # ---------------------------- update critic ---------------------------- #
        # Get predicted next-state actions and Q values from target models
        with torch.no_grad():
            next_action, new_log_pi = self.actor_local.evaluate(next_states)
            Q_target1_next = self.critic1_target(next_states, next_action)
            Q_target2_next = self.critic2_target(next_states, next_action)
            Q_target_next = torch.min(Q_target1_next, Q_target2_next) - self.alpha * new_log_pi
            # Compute Q targets for current states (y_i)
            Q_targets = rewards + (self.gamma * (1 - dones) * Q_target_next)

        # Compute critic loss
        q1 = self.critic1(states, actions)
        q2 = self.critic2(states, actions)

        critic1_loss = F.mse_loss(q1, Q_targets)
        critic2_loss = F.mse_loss(q2, Q_targets)

        # Update critics
        # critic 1
        self.critic1_optimizer.zero_grad()
        critic1_loss.backward(retain_graph=True)
        clip_grad_norm_(self.critic1.parameters(), self.clip_grad_param)
        self.critic1_optimizer.step()
        # critic 2
        self.critic2_optimizer.zero_grad()
        critic2_loss.backward()
        clip_grad_norm_(self.critic2.parameters(), self.clip_grad_param)
        self.critic2_optimizer.step()

        # ---------------------------- update safety critic ---------------------------- #
        # Get predicted next-state actions and Q values from target models
        with torch.no_grad():
            next_action, new_log_pi = self.actor_local.evaluate(next_states)
            Q_target1_next = self.safety_critic1_target(next_states, next_action)
            Q_target2_next = self.safety_critic2_target(next_states, next_action)
            Q_target_next = torch.min(Q_target1_next, Q_target2_next) # - self.alpha * new_log_pi
            # Compute Q targets for current states (y_i)
            Q_targets = costs + (self.gamma * (1 - dones) * Q_target_next)

        # Compute safety_critic loss
        q1 = self.safety_critic1(states, actions)
        q2 = self.safety_critic2(states, actions)

        safety_critic1_loss = F.mse_loss(q1, Q_targets)
        safety_critic2_loss = F.mse_loss(q2, Q_targets)
        
        # CQL addon
        num_repeat = 10
        random_actions = torch.FloatTensor(q1.shape[0] * num_repeat, actions.shape[-1]).uniform_(-1, 1).to(self.device)
        temp_states = states.unsqueeze(1).repeat(1, num_repeat, 1).view(states.shape[0] * num_repeat, states.shape[1])
        temp_next_states = next_states.unsqueeze(1).repeat(1, num_repeat, 1).view(next_states.shape[0] * num_repeat, next_states.shape[1])
        
        current_pi_values1, current_pi_values2  = self._compute_policy_values(temp_states, temp_states)
        next_pi_values1, next_pi_values2 = self._compute_policy_values(temp_next_states, temp_states)
        
        random_values1 = self._compute_random_values(temp_states, random_actions, self.safety_critic1).reshape(states.shape[0], num_repeat, 1)
        random_values2 = self._compute_random_values(temp_states, random_actions, self.safety_critic2).reshape(states.shape[0], num_repeat, 1)
        
        current_pi_values1 = current_pi_values1.reshape(states.shape[0], num_repeat, 1)
        current_pi_values2 = current_pi_values2.reshape(states.shape[0], num_repeat, 1)

        next_pi_values1 = next_pi_values1.reshape(states.shape[0], num_repeat, 1)
        next_pi_values2 = next_pi_values2.reshape(states.shape[0], num_repeat, 1)
        
        cat_q1 = torch.cat([random_values1, current_pi_values1, next_pi_values1], 1)
        cat_q2 = torch.cat([random_values2, current_pi_values2, next_pi_values2], 1)
        
        assert cat_q1.shape == (states.shape[0], 3 * num_repeat, 1), f"cat_q1 instead has shape: {cat_q1.shape}"
        assert cat_q2.shape == (states.shape[0], 3 * num_repeat, 1), f"cat_q2 instead has shape: {cat_q2.shape}"
        
        # flipped sign of cql1_scaled_loss and cql2_scaled_loss
        cql1_scaled_loss = -(torch.logsumexp(cat_q1 / self.cql_temp, dim=1).mean() * self.cql_weight * self.cql_temp) + (q1.mean() * self.cql_weight)
        cql2_scaled_loss = -(torch.logsumexp(cat_q2 / self.cql_temp, dim=1).mean() * self.cql_weight * self.cql_temp) + (q2.mean() * self.cql_weight)
        
        cql_alpha_loss = torch.FloatTensor([0.0])
        cql_alpha = torch.FloatTensor([1.0])
        if self.cql_with_lagrange:
            cql_alpha = torch.clamp(self.cql_log_alpha.exp(), min=0.0, max=1000000.0)
            cql1_scaled_loss = cql_alpha * (cql1_scaled_loss - self.cql_target_action_gap)
            cql2_scaled_loss = cql_alpha * (cql2_scaled_loss - self.cql_target_action_gap)

            self.cql_alpha_optimizer.zero_grad()
            cql_alpha_loss = (- cql1_scaled_loss - cql2_scaled_loss) * 0.5 
            cql_alpha_loss.backward(retain_graph=True)
            self.cql_alpha_optimizer.step()
        
        total_c1_loss = safety_critic1_loss + cql1_scaled_loss
        total_c2_loss = safety_critic2_loss + cql2_scaled_loss
        
        
        # Update safety_critics
        # safety_critic 1
        self.safety_critic1_optimizer.zero_grad()
        total_c1_loss.backward(retain_graph=True)
        clip_grad_norm_(self.safety_critic1.parameters(), self.clip_grad_param)
        self.safety_critic1_optimizer.step()
        # safety_critic 2
        self.safety_critic2_optimizer.zero_grad()
        total_c2_loss.backward()
        clip_grad_norm_(self.safety_critic2.parameters(), self.clip_grad_param)
        self.safety_critic2_optimizer.step()

        # ---------------------------- update actor ---------------------------- #
        # Equation 6 of the paper
        # Estimate reward advantage
        q1 = self.critic1(states, actions)
        q2 = self.critic2(states, actions)
        v = torch.min(q1, q2).detach()

        new_action, new_log_pi = self.actor_local.evaluate(states)
        q1 = self.critic1(states, new_action)
        q2 = self.critic2(states, new_action)
        q = torch.min(q1, q2)

        reward_advantage = q - v

        # Estimate cost advantage
        q1 = self.safety_critic1(states, actions)
        q2 = self.safety_critic2(states, actions)
        v = torch.min(q1, q2).detach()

        q1 = self.safety_critic1(states, new_action)
        q2 = self.safety_critic2(states, new_action)
        q = torch.min(q1, q2)

        cost_advantage = (q - v).mean()

        # Calculate losses
        reward_loss = ((reward_advantage - self.alpha * new_log_pi)).mean()
        cost_loss = self.csc_lambda.detach() / (1 - self.gamma) * cost_advantage
        total_loss = -(reward_loss - cost_loss)
        
        # Calculate objective gradients
        self.actor_optimizer.zero_grad()
        total_loss.backward(retain_graph=True)
        objective_grads = self.flatten_gradients(self.actor_local).reshape((-1, 1))

        # Calculate fisher matrix
        n = objective_grads.numel()
        b = new_log_pi.shape[0]
        fisher = torch.zeros((n,n), dtype=torch.get_default_dtype()).to(self.device)
        for i in range(b):
            self.actor_optimizer.zero_grad()
            new_log_pi[i].backward(retain_graph=bool(i+1<b))
            g = self.flatten_gradients(self.actor_local).reshape((-1, 1))
            fisher += g @ g.T
        fisher /= n

        # Calculate update gradient
        if self.fisher_inversion == "pinv":
            update_grads = torch.linalg.lstsq(fisher.cpu(), objective_grads.cpu()).solution.flatten().to(self.device)
        elif self.fisher_inversion == "inv":
            fisher += 1e-5 * torch.eye(n).to(self.device)   # add some small constant to make it non singular
            update_grads = torch.linalg.solve(fisher, objective_grads).flatten()
        elif self.fisher_inversion == "none":
            update_grads = objective_grads.flatten()

        # Calculate beta terms
        beta_term = torch.sqrt((2*self.csc_delta) / (objective_grads.T @ fisher @ objective_grads))
        beta_j = self.csc_beta

        # Store old parameters
        actor_backup = copy.deepcopy(self.actor_local)
        actor_backup.load_state_dict(self.actor_local.state_dict())

        for j in range(self.csc_update_iterations):
            beta_j *= (1 - beta_j)**j
            beta = (beta_j * beta_term).item()
            
            # Apply gradient with beta as learning rate
            self.apply_flattened_gradient(
                model=self.actor_local,
                grads=update_grads,
                lr=beta
            )

            # Calculate log_probs and kl_div for new actions
            _, next_log_pi = self.actor_local.evaluate(states)
            kl_div = F.kl_div(new_log_pi, next_log_pi, log_target=True, reduction='mean').item()    # in our case same as batchmean

            if kl_div <= self.csc_delta:
                break
            else:
                # Restore original parameters
                self.hard_update(
                    local_model=actor_backup,
                    target_model=self.actor_local,
                )

        # Update alpha
        alpha_loss = - (self.log_alpha.exp() * (new_log_pi + self.target_entropy).detach()).mean()
        self.alpha_optimizer.zero_grad()
        alpha_loss.backward()
        self.alpha_optimizer.step()
        self.alpha = self.log_alpha.exp().detach()

        # ---------------------------- update csc lambda ---------------------------- #
        # Equation 6 of the paper
        # Estimate cost advantage
        with torch.no_grad():
            q1 = self.safety_critic1(states, actions)
            q2 = self.safety_critic2(states, actions)
            v = torch.min(q1, q2)

            new_action, new_log_pi = self.actor_local.evaluate(states)
            q1 = self.safety_critic1(states, new_action)
            q2 = self.safety_critic2(states, new_action)
            q = torch.min(q1, q2)

            cost_advantage = (q - v).mean()

        # Compute csc lambda loss
        csc_lambda_loss = -self.csc_lambda*(self.csc_avg_unsafe + (1 / (1 - self.gamma)) * cost_advantage - self.csc_chi)

        self.csc_lambda_optimizer.zero_grad()
        csc_lambda_loss.backward()
        self.csc_lambda_optimizer.step()
        self.csc_lambda.data.clamp_(min=0.0)

        # ----------------------- update target networks ----------------------- #
        self.soft_update(self.critic1, self.critic1_target, tau=self.tau)
        self.soft_update(self.critic2, self.critic2_target, tau=self.tau)
        self.soft_update(self.safety_critic1, self.safety_critic1_target, tau=self.tau)
        self.soft_update(self.safety_critic2, self.safety_critic2_target, tau=self.tau)
        
        # ----------------------- update stats ----------------------- #
        data = {
            "actor_loss": total_loss.item(),
            "alpha_loss": alpha_loss.item(),
            "alpha": self.alpha.item(),
            "lambda_loss": csc_lambda_loss.item(),
            "lambda": self.csc_lambda.item(),
            "critic1_loss": critic1_loss.item(),
            "critic2_loss": critic2_loss.item(),
            "cql1_scaled_loss": cql1_scaled_loss.item(),
            "cql2_scaled_loss": cql2_scaled_loss.item(),
            "total_c1_loss": total_c1_loss.item(),
            "total_c2_loss": total_c2_loss.item(),
            "cql_alpha_loss": cql_alpha_loss.item(),
            "cql_alpha": cql_alpha.item(),
            "csc_update_iterations": j,
            "csc_kl_div": kl_div,
        }
        if self.stats.total_updates % 8 == 0:
            self.stats.log_update_tensorboard(data)
        self.stats.total_updates += 1
        return data

    def hard_update(self, local_model , target_model):
        self.soft_update(local_model, target_model, tau=1.0)

    def soft_update(self, local_model , target_model, tau=0.005):
        """Soft update model parameters.
        θ_target = τ*θ_local + (1 - τ)*θ_target
        Params
        ======
            local_model: PyTorch model (weights will be copied from)
            target_model: PyTorch model (weights will be copied to)
            tau (float): interpolation parameter 
        """
        for target_param, local_param in zip(target_model.parameters(), local_model.parameters()):
            target_param.data.copy_(tau*local_param.data + (1.0-tau)*target_param.data)
    
    def apply_flattened_gradient(self, model, grads, lr):
        idx = 0
        for name, param in model.named_parameters():
            # Check if this param requires updating the gradient
            if not param.requires_grad: continue
            if param.grad is None: continue
            # Extract gradient and reshape
            n = param.numel()
            g = grads[idx:idx+n].reshape(param.shape)
            # Update parameter and idx
            param.data.copy_(param.data + lr*g)
            idx += n
        assert(idx == grads.numel())
    
    def flatten_gradients(self, model):
        grads = []
        for name, param in model.named_parameters():
            if not param.requires_grad: continue
            if param.grad is None: continue
            grads.append(param.grad.detach().clone().flatten())
        return torch.cat(grads)